using System.Collections.Generic;

using System;



public class Parser {
	const int _EOF = 0;
	const int _IntegerNumber = 1;
	const int maxT = 11;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;

	public static Token t;    // last recognized token
	public static Token la;   // lookahead token
	static int errDist = minErrDist;

public static List<int> program = new List<int>();



	static void SynErr (int n) {
		if (errDist >= minErrDist) Errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public static void SemErr (string msg) {
		if (errDist >= minErrDist) Errors.Error(t.line, t.col, msg);
		errDist = 0;
	}
	
	static void Get () {
		for (;;) {
			t = la;
			la = Scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }

			la = t;
		}
	}
	
	static void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	static bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	static void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	static bool WeakSeparator (int n, int syFol, int repFol) {
		bool[] s = new bool[maxT+1];
		if (la.kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			for (int i=0; i <= maxT; i++) {
				s[i] = set[syFol, i] || set[repFol, i] || set[0, i];
			}
			SynErr(n);
			while (!s[la.kind]) Get();
			return StartOf(syFol);
		}
	}
	
	static void xP() {
		Expr();
		program.Add(7); 
		while (la.kind == 1 || la.kind == 2 || la.kind == 9) {
			Expr();
			program.Add(7); 
		}
		program.Add(6); 
	}

	static void Expr() {
		int bp1 = 0, bp2 = 0; 
		if (la.kind == 2) {
			Get();
			Expr();
			program.Add(5); bp1 = program.Count; program.Add(0); 
			Expect(3);
			Expr();
			program.Add(0); program.Add(0); program.Add(5); bp2 = program.Count; program.Add(0); program[bp1] = program.Count; 
			Expect(4);
			Expr();
			program[bp2] = program.Count; 
		} else if (la.kind == 1 || la.kind == 9) {
			SubMulDivTerm();
			if (la.kind == 5) {
				Get();
				Expr();
				program.Add(1); 
			}
		} else SynErr(12);
	}

	static void SubMulDivTerm() {
		MulDivTerm();
		if (la.kind == 6) {
			Get();
			SubMulDivTerm();
			program.Add(2); 
		}
	}

	static void MulDivTerm() {
		DivTerm();
		if (la.kind == 7) {
			Get();
			MulDivTerm();
			program.Add(3); 
		}
	}

	static void DivTerm() {
		Term();
		if (la.kind == 8) {
			Get();
			DivTerm();
			program.Add(4); 
		}
	}

	static void Term() {
		if (la.kind == 1) {
			Get();
			program.Add(0); program.Add(int.Parse(t.val)); 
		} else if (la.kind == 9) {
			Get();
			Expr();
			Expect(10);
		} else SynErr(13);
	}



	public static void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		xP();

    Expect(0);
	}

	static bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x}

	};
} // end Parser


public class Errors {
	public static int count = 0;                                    // number of errors detected
  public static string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
	
	public static void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "IntegerNumber expected"; break;
			case 2: s = "\"[\" expected"; break;
			case 3: s = "\"]\" expected"; break;
			case 4: s = "\":\" expected"; break;
			case 5: s = "\"+\" expected"; break;
			case 6: s = "\"-\" expected"; break;
			case 7: s = "\"*\" expected"; break;
			case 8: s = "\"/\" expected"; break;
			case 9: s = "\"(\" expected"; break;
			case 10: s = "\")\" expected"; break;
			case 11: s = "??? expected"; break;
			case 12: s = "invalid Expr"; break;
			case 13: s = "invalid Term"; break;

			default: s = "error " + n; break;
		}
		Console.WriteLine(Errors.errMsgFormat, line, col, s);
		count++;
	}

	public static void SemErr (int line, int col, int n) {
		Console.WriteLine(errMsgFormat, line, col, ("error " + n));
		count++;
	}

	public static void Error (int line, int col, string s) {
		Console.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public static void Exception (string s) {
		Console.WriteLine(s); 
		System.Environment.Exit(1);
	}
} // Errors

