﻿#region Using directives

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace PA {
  class SingleLinkedList {
    private SingleLinkedList next;
    private int val;
    public SingleLinkedList Next {
      get { return next; }
      set { next = value; }
    }
    public int Value {
      get { return val; }
      set { val = value; }
    }
    public SingleLinkedList(int v, SingleLinkedList n) {
      val = v;
      next = n;
    }
    public SingleLinkedList(int v) : this(v, null) { }
    public SingleLinkedList() : this(0, null) { }
    public SingleLinkedList(int[] data) {
      SingleLinkedList p = this;
      val = data[0];
      next = null;
      for (int i = 1; i < data.Length; i++) {
        p.Next = new SingleLinkedList(data[i]);
        p = p.Next;
      }
    }

    public static SingleLinkedList Common(SingleLinkedList[] lists) {
      SingleLinkedList ret = null, tail = null;
      SingleLinkedList[] cursors = (SingleLinkedList[])lists.Clone();
      int cur = 0;
      int start = -1;

      while (true) {
        int next = (cur + 1) % lists.Length;
        if (cursors[cur] == null || cursors[next] == null)
          return ret;
        if (cursors[cur].val < cursors[next].val) {
          start = -1;
          while (cursors[cur] != null && cursors[cur].val < cursors[next].val)
            cursors[cur] = cursors[cur].Next;
          continue;
        } else if (start == -1 && cursors[cur].val == cursors[next].val) {
          start = cur;
        } else if (next == start) {
          start = -1;
          if (ret == null)
            tail = ret = new SingleLinkedList(cursors[cur].val);
          else if (tail.val != cursors[cur].val)
            tail.Next = new SingleLinkedList(cursors[cur].val);
          cursors[next] = cursors[next].Next;
        }
        cur = next;
      }
    }
  }

  class Interpreter {
    enum OpCodes {
      ldc_i4 = 0,
      add	= 1,
      sub =	2,
      mul	= 3,
      div	= 4,
      brzero = 5,
      ret	= 6,
      print	= 7,
      dup	= 8
    }
    Stack<int> stack;
    int pc;
    int[] memory;
    public Interpreter(int[] program) {
      memory = program;
    }
    public void Run() {
      stack = new Stack<int>();
      pc = 0;
      int tmp;
      while (true) {
        switch ((OpCodes)memory[pc++]) {
          case OpCodes.ldc_i4:
            stack.Push(memory[pc++]);
            break;
          case OpCodes.add:
            stack.Push(stack.Pop() + stack.Pop());
            break;
          case OpCodes.sub:
            tmp = stack.Pop();
            stack.Push(stack.Pop() - tmp);
            break;
          case OpCodes.mul:
            stack.Push(stack.Pop() * stack.Pop());
            break;
          case OpCodes.div:
            tmp = stack.Pop();
            stack.Push(stack.Pop() / tmp);
            break;
          case OpCodes.brzero:
            if (stack.Pop() == 0)
              pc = memory[pc];
            else
              pc++;
            break;
          case OpCodes.ret:
            Console.WriteLine("Exiting...");
            return;
          case OpCodes.print:
            Console.WriteLine(stack.Pop());
            break;
          case OpCodes.dup:
            tmp = stack.Pop();
            stack.Push(tmp);
            stack.Push(tmp);
            break;
          default:
            throw new Exception("Unknown opcode");
        }
      }
    }
  }

  class Program {
    static void testEx1() {
      SingleLinkedList[] l = new SingleLinkedList[3];
      l[0] = new SingleLinkedList(new int[] { 3, 5, 10, 20 });
      l[1] = new SingleLinkedList(new int[] { 10, 11, 12, 20 });
      l[2] = new SingleLinkedList(new int[] { 3, 10, 12, 20 });

      SingleLinkedList p = SingleLinkedList.Common(l);
      while (p != null) {
        Console.WriteLine(p.Value);
        p = p.Next;
      }
    }

    static void testEx2() {
      Interpreter interp = new Interpreter(new int[]{0, 1, 0, 2, 1, 8, 7, 0, 1, 2, 7, 6});
      interp.Run();
    }

    static void testEx3() {
      Scanner.Init("test.xp");
      Parser.Parse();
      Interpreter interp = new Interpreter(Parser.program.ToArray());
      interp.Run();
    }

    static void Main(string[] args) {
      testEx3();
    }
  }
}
